{ ****************************************************************************** }
{                                                                                }
{  Vertex buffer object demo -  2003 by Sascha Willems                          }
{   visit http://www.delphigl.de for more                                        }
{   visit http://www.delphigl.com the biggest german Delphi+OpenGL community     }
{                                                                                }
{  More infos on GL_ARB_VERTEX_BUFFER_OBJECT can be found here :                 }
{   http://oss.sgi.com/projects/ogl-sample/registry/ARB/vertex_buffer_object.txt }
{                                                                                }
{ ****************************************************************************** }
{  Note :                                                                        }
{   In order to run this demo, you'll need a 3D-accelerator that supports the    }
{   GL_ARB_Vertex_Buffer_Object-extension.                                       }
{ ****************************************************************************** }
{  Compiled with Delphi 7.0                                                      }
{ ****************************************************************************** }

unit GL_ARB_Vertex_Buffer_Object1;

interface

uses
  Windows,
  Messages,
  SysUtils,
  Variants,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  AppEvnts,

  dglOpenGL,
  glMisc,
  textures,
  ClipBrd,

  ExtCtrls;

type
  TGLForm = class(TForm)
    ApplicationEvents: TApplicationEvents;
    FPSTimer: TTimer;
    RotTimer: TTimer;
    procedure ApplicationEventsIdle(Sender: TObject; var Done: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FPSTimerTimer(Sender: TObject);
    procedure RotTimerTimer(Sender: TObject);
    procedure FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
   private
     { Private-Deklarationen }
   public
     { Public-Deklarationen }
  end;
  // -> GL_T2F_V3F
  TVertex = record
    S,T,X,Y,Z : TGLFloat;
   end;

var
  GLForm     : TGLForm;
  DC         : HDC;
  RC         : HGLRC;
  Frames     : Integer;
  Tex        : TGLUInt;
  VBuffer    : TGLUInt;
  Rot        : Single;
  Zoom       : Single;
  FPS        : Integer;
  VBOSize    : Integer;
  VBOPointer : Pointer;

implementation

{$R *.dfm}

procedure CreateHeightMap(pFileName : String;pResolution : Integer);
const
 VertPos : array[0..3] of array[0..1] of Byte = ((0,0),(0,1),(1,1),(1,0));
var
 HeightData : array of array of Integer;
 Vertex     : array[0..3] of array[0..2] of Single;
 Bmp        : TBitMap;
 x,z,i      : LongInt;
 Resolution : Integer;
 Size       : Integer;
 Ratio      : Single;
 VPointer   : ^TVertex;
begin
if not FileExists(pFileName) then
 begin
 MessageDlg('Datei nicht gefunden!', mtError, [mbOK], 0);
 exit;
 end;
Resolution := pResolution;
Size       := 32;
Bmp := TBitMap.Create;
try
 Bmp.LoadFromFile(pFileName);
 Ratio := Bmp.Width/Bmp.Height;
 SetLength(HeightData, Resolution+1);
 for x := 0 to Resolution do
  begin
  SetLength(HeightData[x], Resolution+1);
  for z := 0 to Resolution do
   HeightData[x,z] := Bmp.Canvas.Pixels[Trunc(X/Resolution*Bmp.Width),Trunc(Z/Resolution*Bmp.Height)] div 10000;
  end;
finally
 Bmp.Free
end;
// Bind VBO
glBindBufferARB(GL_ARRAY_BUFFER_ARB, VBuffer);
// Set VBOs size
glBufferDataARB(GL_ARRAY_BUFFER_ARB, Sqr(Resolution-1)*4*SizeOf(TVertex), nil, GL_STATIC_DRAW_ARB);
VBOPointer := glMapBufferARB(GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY_ARB);
VBOSize    := 0;

// Write the vertices of the heightmap directly to VRAM
for x := 0 to Resolution-2 do
 for z := 0 to Resolution-2 do
  begin
  for i := 0 to 3 do
   begin
   Vertex[i,0] := (X + VertPos[I][0])/Resolution*Size-Size/2*Ratio;
   Vertex[i,2] := (Z + VertPos[I][1])/Resolution*Size-Size/2;
   Vertex[i,1] := HeightData[X + VertPos[I][0],Z + VertPos[I][1]]*4;

   VPointer    := VBOPointer;
   VPointer^.X := Vertex[i,0];
   VPointer^.Y := Vertex[i,1]*0.0015;
   VPointer^.Z := Vertex[i,2];
   VPointer^.S := (Vertex[i,0]-Size*Ratio/2)/(Size*Ratio);
   VPointer^.T := (-Vertex[i,2]-Size/2)/Size;

   inc(Integer(VBOPointer), SizeOf(TVertex));
   inc(VBOSize);
   end;
  end;

// We need to unmap our VBO before we can draw it
glUnMapBufferARB(GL_ARRAY_BUFFER_ARB);
end;

procedure TGLForm.FormCreate(Sender: TObject);
begin
Zoom := 50;
InitOpenGL;
DC := GetDC(Handle);
RC := CreateRenderingContext(DC, [opDoubleBuffered], 32, 24, 0, 0, 0, 0);
ActivateRenderingContext(DC, RC);
ReadExtensions;
glClearColor(0,0.5,0.8,0);
glEnable(GL_DEPTH_TEST);
glEnable(GL_TEXTURE_2D);
glDepthFunc(GL_LESS);
LoadTexture('terrain.jpg', Tex, False, GL_LINEAR, GL_LINEAR_MIPMAP_LINEAR, False);
if not glIsExtSupported('GL_ARB_VERTEX_BUFFER_OBJECT') then
 begin
 MessageDlg('GL_ARB_VERTEX_BUFFER_OBJECT not supported!', mtError, [mbOk], 0);
 Application.Terminate;
 end;
// Create VBO
glGenBuffersARB(1, @VBuffer);
CreateHeightMap('hm1k.bmp', 256);
// Set VBO's data
glBindBufferARB(GL_ARRAY_BUFFER_ARB, VBuffer);
ClipBoard.SetTextBuf(PChar(IntToStr(VBOSize*2*9)));
glEnableClientState(GL_VERTEX_ARRAY);
end;

procedure TGLForm.FormDestroy(Sender: TObject);
begin
// Delete VBO
glDeleteBuffersARB(1, @VBuffer);
DeactivateRenderingContext;
DestroyRenderingContext(RC);
ReleaseDC(Handle, DC);
end;

procedure TGLForm.ApplicationEventsIdle(Sender: TObject; var Done: Boolean);
const
 Trans : array[0..8] of array[0..2] of TGLFloat = ( (-32,0,-32), (0,0,-32), (31.5,0,-32),
                                                    (-32,0,0), (0,0,0), (31.5,0,0),
                                                    (-32,0,31.5), (0,0,31.5), (31.5,0,31.5) );
 Scale : array[0..8] of array[0..2] of TGLFloat = ( (-1,1,-1), (1,1,-1), (-1,1,-1),
                                                    (-1,1,1), (1,1,1), (-1,1,1),
                                                    (-1,1,-1), (1,1,-1), (-1,1,-1) );
 PType = GL_QUADS;
var
 i : Integer;
begin
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
glViewport(0,0,ClientWidth,ClientHeight);
gluPerspective(60, ClientWidth/ClientHeight, 0.1, 256);

glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

glTranslatef(0,0,-Zoom);
glRotatef(35, 1, 0, 0);

// Draw VBO(s)
glBindBufferARB(GL_ARRAY_BUFFER_ARB, VBuffer);
for i := Low(Trans) to High(Trans) do
 begin
 glPushMatrix;
  glTranslatef(Trans[i][0], Trans[i][1], Trans[i][2]);
  glScalef(Scale[i][0], Scale[i][1], Scale[i][2]);
  // Our "Vertexarray" (namely the VBO) contains vertices made of 2 Texcoords and 3 vertexcoords (GL_T2F_V3F)
  glInterleavedArrays(GL_T2F_V3F, SizeOf(TVertex), nil);
  glDrawArrays(PType, 0, VBOSize);
 glPopMatrix;
 end;

SwapBuffers(DC);
inc(Frames);
Done := False;
end;

procedure TGLForm.FPSTimerTimer(Sender: TObject);
begin
Caption := 'GL_ARB_Vertex_Buffer_Object ['+IntToStr(Frames)+'fps / '+FloatToStrF((VBOSize*2*9*Frames)/1024/1024, ffFixed, 3, 1)+'MTris/s]';
FPS     := Frames;
Frames  := 0;
end;

procedure TGLForm.RotTimerTimer(Sender: TObject);
begin
Rot := Rot+0.1;
if Rot > 360 then
 Rot := 0;
end;

procedure TGLForm.FormMouseWheelDown(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
begin
Zoom := Zoom+0.1;
end;

procedure TGLForm.FormMouseWheelUp(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
begin
Zoom := Zoom-0.1;
end;

procedure TGLForm.FormKeyPress(Sender: TObject; var Key: Char);
var
 BuffInfo : Integer;
begin
if Key = 'i' then
 begin
 glGetBufferParameterivARB(GL_ARRAY_BUFFER_ARB, GL_BUFFER_SIZE_ARB, @BuffInfo);
 ShowMessage('VBO-Size  = '+IntToStr(BuffInfo));
 end;
end;

end.
