unit glMisc;

interface

uses
 SysUtils,
 JPEG,
 Windows,
 Graphics,
 dglOpenGL;

type
 PRGBQuad = ^TRGBQuad;
 TRGBQuad = packed record
   Red   : Byte;
   Green : Byte;
   Blue  : Byte;
   Alpha : Byte;
  end;
 TMatrix16f = array[0..15] of Single;
 TVector4f  = array[0..3] of Single;

function glGetErrorStr : String;
function glIsExtSupported(fExtName : String) : Boolean;
procedure glSaveScreen(pFilename : String);

function CalculateShadowMatrix(fLightPos : TVector4f;fPlane : TVector4f) : TMatrix16f;

implementation

// =============================================================================
//  glGetErrorStr
// =============================================================================
//  Gibt den letzten OpenGL-Fehler als String zurck
// =============================================================================
function glGetErrorStr : String;
var
 glError : TGLUInt;
begin
glError := glGetError;
case glError of
 GL_NO_ERROR          : Result := 'GL_NO_ERROR';
 GL_INVALID_ENUM      : Result := 'GL_INVALID_ENUM';
 GL_INVALID_VALUE     : Result := 'GL_INVALID_VALUE';
 GL_INVALID_OPERATION : Result := 'GL_INVALID_OPERATION';
 GL_STACK_OVERFLOW    : Result := 'GL_STACK_OVERFLOW';
 GL_STACK_UNDERFLOW   : Result := 'GL_STACK_UNDERFLOW';
 GL_OUT_OF_MEMORY     : Result := 'GL_OUT_OF_MEMORY';
end;
end;

// =============================================================================
//  glIsExtSupported
// =============================================================================
//  Gibt TRUE zurck, falls die angefragte Extension untersttzt wird
// =============================================================================
function glIsExtSupported(fExtName : String) : Boolean;
var
 Ext       : array[1..512] of String;
 ExtString : String;
 ExtT      : String;
 i,k       : Integer;
begin
ExtString := glGetString(GL_EXTENSIONS);
k := 1;
for i := 1 to Length(ExtString) do
 begin
 if ExtString[i] = ' ' then
  begin
  Ext[k] := ExtT;
  inc(k);
  ExtT := '';
  end
 else
  ExtT := ExtT+ExtString[i];
 end;
Result := false;
for i := 1 to 512 do
 if UpperCase(Ext[i]) = UpperCase(fExtName) then
  Result := true;
end;

// =============================================================================
//  glSaveScreen
// =============================================================================
//  Speichert einen Screenshot des aktuellen Pufferinhaltes
// =============================================================================
procedure glSaveScreen(pFilename : String);
var
 Viewport : array[0..3] of TGLint;
 JPG      : TJPEGImage;
 RGBBits  : PRGBQuad;
 Pixel    : PRGBQuad;
 BMP      : TBitmap;
 Header   : PBitmapInfo;
 x,y      : Integer;
 Temp     : Byte;
begin
glGetIntegerv(GL_VIEWPORT, @Viewport);
GetMem(RGBBits, Viewport[2]*Viewport[3]*4);
glFinish;
glPixelStorei(GL_PACK_ALIGNMENT, 4);
glPixelStorei(GL_PACK_ROW_LENGTH, 0);
glPixelStorei(GL_PACK_SKIP_ROWS, 0);
glPixelStorei(GL_PACK_SKIP_PIXELS, 0);
glReadPixels(0, 0, Viewport[2], Viewport[3], GL_RGBA, GL_UNSIGNED_BYTE, RGBBits);
// Screenshot als JPG speichern
JPG := TJPEGImage.Create;
BMP := TBitmap.Create;
BMP.PixelFormat := pf32Bit;
BMP.Width       := Viewport[2];
BMP.Height      := Viewport[3];
GetMem(Header, SizeOf(TBitmapInfoHeader));
with Header^.bmiHeader do
 begin
 biSize        := SizeOf(TBitmapInfoHeader);
 biWidth       := Viewport[2];
 biHeight      := Viewport[3];
 biPlanes      := 1;
 biBitCount    := 32;
 biCompression := BI_RGB;
 biSizeImage   := Viewport[2]*Viewport[3]*4;
 end;
// Rot und Blau vertauschen
Pixel := RGBBits;
for x := 0 to Viewport[2]-1 do
 for y := 0 to Viewport[3]-1 do
  begin
  Temp       := Pixel.Red;
  Pixel.Red  := Pixel.Blue;
  Pixel.Blue := Temp;
  inc(Pixel);
  end;
SetDIBits(Bmp.Canvas.Handle, Bmp.Handle, 0, Viewport[3], RGBBits, TBitmapInfo(Header^), DIB_RGB_COLORS);
JPG.CompressionQuality := 100;
JPG.Assign(BMP);
JPG.SaveToFile(pFileName);
FreeMem(Header);
FreeMem(RGBBits);
JPG.Free;
BMP.Free;
end;

// =============================================================================
//  CalculateShadowMatrix
// =============================================================================
function CalculateShadowMatrix(fLightPos : TVector4f;fPlane : TVector4f) : TMatrix16f;
var
 Dot : Single;
begin
// Get the dot product of the light and the plane vectors
Dot := fPlane[0]*fLightPos[0] + fPlane[1]*fLightPos[1] + fPlane[2]*fLightPos[2] + fPlane[3]*fLightPos[3];
// First column
Result[ 0] := Dot - fLightPos[0] * fPlane[0];
Result[ 4] :=   0 - fLightPos[0] * fPlane[1];
Result[ 8] :=   0 - fLightPos[0] * fPlane[2];
Result[12] :=   0 - fLightPos[0] * fPlane[3];
// second column
Result[1]  :=   0 - fLightPos[1] * fPlane[0];
Result[5]  := Dot - fLightPos[1] * fPlane[1];
Result[9]  :=   0 - fLightPos[1] * fPlane[2];
Result[13] :=   0 - fLightPos[1] * fPlane[3];
// third column
Result[2]  :=   0 - fLightPos[2] * fPlane[0];
Result[6]  :=   0 - fLightPos[2] * fPlane[1];
Result[10] := Dot - fLightPos[2] * fPlane[2];
Result[14] :=   0 - fLightPos[2] * fPlane[3];
// fourth column
Result[3]  :=   0 - fLightPos[3] * fPlane[0];
Result[7]  :=   0 - fLightPos[3] * fPlane[1];
Result[11] :=   0 - fLightPos[3] * fPlane[2];
Result[15] := Dot - fLightPos[3] * fPlane[3];
end;

end.
